use std::fmt::Display;

use crate::{
    class::Class,
    combat::EncounterType,
    equipment::{Armor, Weapon, WeaponType},
    skills::{Proficiencies, ToProficiencies},
    stats::{Stat, StatBlock, ToStatBlock},
};

#[derive(Debug, Clone)]
pub struct Character {
    pub name: String,
    pub class1: Class,
    pub class2: Class,
    pub stat_block: StatBlock,
    pub hp: i64,
    pub mp: i64,
    pub proficiency: i64,
    pub proficiencies: Proficiencies,
    pub armor: Armor,
    pub weapon: Weapon,
    dice_pool: i64,
}

impl Character {
    pub fn ac(&self) -> i64 {
        self.stat(Stat::STB) + self.stat(Stat::DEX) + self.armor_value()
    }

    pub fn weapon_proficiency(&self) -> i64 {
        if match self.weapon.weapon_type {
            WeaponType::SimpleWeapon => self.proficiencies.simple_weapons,
            WeaponType::BladedWeapon => self.proficiencies.bladed_weapons,
            WeaponType::BluntWeapon => self.proficiencies.blunt_weapons,
            WeaponType::PoleWeapon => self.proficiencies.pole_weapons,
            WeaponType::ThrownWeapon => self.proficiencies.thrown_weapons,
            WeaponType::RangedWeapon => self.proficiencies.ranged_weapons,
        } {
            self.proficiency
        } else {
            0
        }
    }

    pub fn armor_value(&self) -> i64 {
        if match self.armor {
            Armor::Heavy => self.proficiencies.heavy_armor,
            Armor::Medium => self.proficiencies.medium_armor,
            Armor::Light => self.proficiencies.light_armor,
            Armor::None => true,
        } {
            self.armor.armor()
        } else {
            0
        }
    }

    pub fn compute_damage(&self, dice: i64) -> i64 {
        (dice + self.weapon_proficiency()).max(0) * self.weapon.damage_dice
            + if let Some(damage_stat) = self.weapon.stat_damage {
                self.stat(damage_stat)
            } else {
                0
            }
    }

    pub fn dice_pool(&self) -> i64 {
        self.dice_pool
    }

    pub fn init_dice_pool(&mut self, encounter_type: EncounterType) {
        self.dice_pool = match encounter_type {
            EncounterType::Physical => {
                self.stat_block[self.weapon.weapon_stat] + self.weapon_proficiency()
            }
            EncounterType::Mental => self.stat_block.INT,
        }
        .max(0)
    }

    pub fn adjust_dice_pool(&mut self, adjustment: i64, encounter_type: EncounterType) {
        self.dice_pool = (self.dice_pool
            + adjustment
            + match encounter_type {
                EncounterType::Physical => self.stat_block.CON,
                EncounterType::Mental => self.stat_block.CMP,
            })
        .max(0)
    }

    pub fn init_hp_mp(&mut self, encounter_type: EncounterType) {
        if encounter_type == EncounterType::Physical {
            self.hp = (self.stat(Stat::CON) + self.stat(Stat::STB)) * (2 + self.proficiency);
        } else {
            self.mp = (self.stat(Stat::CMP) + self.stat(Stat::CHA)) * (2 + self.proficiency);
        }
    }

    pub fn init_encounter(&mut self, encounter_type: EncounterType) {
        self.init_dice_pool(encounter_type);
        self.init_hp_mp(encounter_type);
    }

    pub fn stat(&self, stat: Stat) -> i64 {
        let adjusted_stat_block = self.stat_block + self.armor.stat_adjust();

        adjusted_stat_block[stat]
    }
}

pub fn create_character(
    name: String,
    class1: Class,
    class2: Class,
    stat_choice: StatBlock,
    stat_adjust: StatBlock,
    armor: Armor,
    weapon: Weapon,
) -> Character {
    if !(Vec::from(stat_choice).iter().all(|s| *s >= 0)
        && Vec::from(stat_choice).iter().sum::<i64>() == 2i64)
    {
        panic!("Invalid stat choice: {:?}", stat_choice)
    }
    if Vec::from(stat_adjust).iter().sum::<i64>() != 0 {
        panic!("Invalid stat adjust: {:?}", stat_adjust)
    }
    let stat_block =
        StatBlock::ones() + class1.stat_block() + class2.stat_block() + stat_choice + stat_adjust;
    Character {
        name,
        class1,
        class2,
        stat_block,
        hp: (stat_block.CON + stat_block.STB) * 3,
        mp: (stat_block.CMP + stat_block.CHA) * 3,
        proficiency: 1,
        proficiencies: class1.proficiencies() | class2.proficiencies(),
        armor,
        weapon,
        dice_pool: 0,
    }
}

impl Display for Character {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(
            f,
            r#"╔═══════════════════════════════╤═══════╗
║ {name:<15}  {class1:>12} │ HP {hp:>2} ║
║                  {class2:>12} │ MP {mp:>2} ║
╠═══════════════════╤═══════════╧═══════╣
║            STR {str:>2} │ {int:>2} INT            ║
║            DEX {dex:>2} │ {knd:>2} KND            ║
║            CON {con:>2} │ {cmp:>2} CMP            ║
║            STB {stb:>2} │ {cha:>2} CHA            ║
╠═══════════════════╧═══════════════════╣
║  Armor  {armor:<15}         AC {ac:>2} ║
║ Weapon  {weapon:<15}               ║
╠═══════════════════════════════════════╣
║ Dice pool {dp:>2}                          ║
╚═══════════════════════════════════════╝"#,
            name = self.name,
            class1 = self.class1,
            class2 = self.class2,
            hp = self.hp,
            mp = self.mp,
            ac = self.ac(),
            str = self.stat_block.STR,
            dex = self.stat_block.DEX,
            con = self.stat_block.CON,
            stb = self.stat_block.STB,
            int = self.stat_block.INT,
            knd = self.stat_block.KND,
            cmp = self.stat_block.CMP,
            cha = self.stat_block.CHA,
            armor = self.armor,
            weapon = self.weapon,
            dp = self.dice_pool()
        )
    }
}
