use rand::Rng;
use std::ops;

#[derive(Debug, Clone, Copy)]
#[allow(non_camel_case_types)]
pub enum Dice {
    d4,
    d6,
    d8,
    d10,
    d12,
    d20,
    d100,
}

impl Iterator for Dice {
    type Item = i64;

    fn next(self: &mut Dice) -> Option<Self::Item> {
        let mut rng = rand::thread_rng();
        match self {
            Dice::d4 => Some(rng.gen_range(1..4)),
            Dice::d6 => Some(rng.gen_range(1..6)),
            Dice::d8 => Some(rng.gen_range(1..8)),
            Dice::d10 => Some(rng.gen_range(1..10)),
            Dice::d12 => Some(rng.gen_range(1..12)),
            Dice::d20 => Some(rng.gen_range(1..20)),
            Dice::d100 => Some(rng.gen_range(1..100)),
        }
    }
}

impl ops::Mul<Dice> for i64 {
    type Output = i64;

    fn mul(self, rhs: Dice) -> Self::Output {
        rhs.take(self as usize).sum()
    }
}

pub fn successes(n: usize, threshold: i64) -> (i64, i64, i64, i64, i64) {
    if threshold < 1 || threshold > 12 {
        panic!("Invalid threshold: {}", threshold)
    }

    let rolls: Vec<i64> = Dice::d12.take(n).collect();

    let succ_dice = rolls.iter().filter(|roll| **roll > threshold).count() as i64;
    let crit_succ = rolls.iter().filter(|roll| **roll == 12).count() as i64;

    let fail_dice = rolls.iter().filter(|roll| **roll <= threshold).count() as i64;
    let crit_fail = rolls.iter().filter(|roll| **roll == 1).count() as i64;

    (
        succ_dice + crit_succ - fail_dice - crit_fail,
        succ_dice + crit_succ,
        fail_dice + crit_fail,
        succ_dice,
        fail_dice,
    )
}

#[cfg(test)]
mod dice_tests {

    use super::*;

    #[test]
    fn in_range() {
        assert!(Dice::d4.take(1000).all(|roll| roll >= 1 && roll <= 4));
        assert!(Dice::d6.take(1000).all(|roll| roll >= 1 && roll <= 6));
        assert!(Dice::d8.take(1000).all(|roll| roll >= 1 && roll <= 8));
        assert!(Dice::d10.take(1000).all(|roll| roll >= 1 && roll <= 10));
        assert!(Dice::d12.take(1000).all(|roll| roll >= 1 && roll <= 12));
        assert!(Dice::d20.take(1000).all(|roll| roll >= 1 && roll <= 20));
        assert!(Dice::d100.take(1000).all(|roll| roll >= 1 && roll <= 100))
    }

    #[test]
    #[should_panic]
    fn wrong_threshold() {
        successes(10, 0);
    }
}

pub trait Clampable: Ord + Copy {
    fn clamp(&self, lb: &Self, ub: &Self) -> Self {
        *self.min(ub).max(lb)
    }
}

impl Clampable for i64 {}
