use std::fmt::Display;

use crate::{
    dice::Dice,
    stats::{Stat, StatBlock},
};
use rand::seq::IteratorRandom;
use rand_distr::{Distribution, Standard};
use strum::IntoEnumIterator;
use strum_macros::{Display, EnumIter};

#[derive(Debug, Clone, Copy, EnumIter, Display)]
pub enum Armor {
    Heavy,
    Medium,
    Light,
    None,
}

impl Armor {
    pub fn armor(&self) -> i64 {
        match self {
            Armor::Heavy => 5,
            Armor::Medium => 3,
            Armor::Light => 1,
            Armor::None => 0,
        }
    }

    pub fn stat_adjust(&self) -> StatBlock {
        match self {
            Armor::Heavy => StatBlock {
                DEX: -1,
                ..Default::default()
            },
            Armor::Medium => Default::default(),
            Armor::Light => Default::default(),
            Armor::None => Default::default(),
        }
    }
}

impl Distribution<Armor> for Standard {
    fn sample<R: rand::Rng + ?Sized>(&self, rng: &mut R) -> Armor {
        Armor::iter().choose(rng).unwrap()
    }
}

#[derive(Debug, Clone, Copy, EnumIter)]
pub enum WeaponType {
    SimpleWeapon,
    BladedWeapon,
    BluntWeapon,
    PoleWeapon,
    ThrownWeapon,
    RangedWeapon,
}

impl WeaponType {
    pub fn create_weapon(&self, name: String) -> Weapon {
        Weapon {
            name,
            weapon_type: *self,
            damage_dice: self.damage_dice(),
            weapon_stat: self.weapon_stat(),
            stat_damage: self.stat_damage(),
        }
    }
    pub fn damage_dice(&self) -> Dice {
        match self {
            WeaponType::SimpleWeapon => Dice::d4,
            WeaponType::BladedWeapon => Dice::d6,
            WeaponType::BluntWeapon => Dice::d10,
            WeaponType::PoleWeapon => Dice::d4,
            WeaponType::ThrownWeapon => Dice::d4,
            WeaponType::RangedWeapon => Dice::d4,
        }
    }

    pub fn weapon_stat(&self) -> Stat {
        match self {
            WeaponType::SimpleWeapon => Stat::STR,
            WeaponType::BladedWeapon => Stat::STR,
            WeaponType::BluntWeapon => Stat::STR,
            WeaponType::PoleWeapon => Stat::DEX,
            WeaponType::ThrownWeapon => Stat::DEX,
            WeaponType::RangedWeapon => Stat::DEX,
        }
    }

    pub fn stat_damage(&self) -> Option<Stat> {
        match self {
            WeaponType::SimpleWeapon => None,
            WeaponType::BladedWeapon => None,
            WeaponType::BluntWeapon => Some(Stat::STR),
            WeaponType::PoleWeapon => Some(Stat::DEX),
            WeaponType::ThrownWeapon => Some(Stat::STR),
            WeaponType::RangedWeapon => None,
        }
    }

    pub fn stat_adjust(&self) -> StatBlock {
        match self {
            WeaponType::SimpleWeapon => StatBlock::default(),
            WeaponType::BladedWeapon => StatBlock::default(),
            WeaponType::BluntWeapon => StatBlock {
                DEX: -1,
                ..Default::default()
            },
            WeaponType::PoleWeapon => StatBlock::default(),
            WeaponType::ThrownWeapon => StatBlock::default(),
            WeaponType::RangedWeapon => StatBlock::default(),
        }
    }
}

impl Display for WeaponType {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            WeaponType::SimpleWeapon => write!(f, "Simple Weapon"),
            WeaponType::BladedWeapon => write!(f, "Bladed Weapon"),
            WeaponType::BluntWeapon => write!(f, "Blunt Weapon"),
            WeaponType::PoleWeapon => write!(f, "Pole Weapon"),
            WeaponType::ThrownWeapon => write!(f, "Thrown Weapon"),
            WeaponType::RangedWeapon => write!(f, "Ranged Weapon"),
        }
    }
}

#[derive(Debug, Clone)]
pub struct Weapon {
    pub name: String,
    pub weapon_type: WeaponType,
    pub damage_dice: Dice,
    pub weapon_stat: Stat,
    pub stat_damage: Option<Stat>,
}

impl Display for Weapon {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.pad(&self.name)
    }
}

impl Distribution<Weapon> for Standard {
    fn sample<R: rand::Rng + ?Sized>(&self, rng: &mut R) -> Weapon {
        let wt = WeaponType::iter().choose(rng).unwrap();
        wt.create_weapon(wt.to_string())
    }
}
