use rand::{
    seq::{IteratorRandom, SliceRandom},
    thread_rng, Rng,
};
use rand_distr::{Distribution, Poisson, Standard};
use strum::IntoEnumIterator;

use crate::{
    character::{create_character, Character},
    stats::{Stat, StatBlock},
};

pub fn random_stat_choice() -> StatBlock {
    let mut rng = thread_rng();

    let stat1: Stat = rng.gen();
    let stat2: Stat = rng.gen();

    stat1 + stat2
}

pub fn random_stat_adjust() -> StatBlock {
    let mut rng = thread_rng();

    let total_adjusts: usize = (Poisson::new(2.0).unwrap().sample(&mut rng) as usize).min(7);
    let n_adjust_stats = rng.gen_range(1..=4);
    let adjust_stats = Stat::iter().choose_multiple(&mut rng, n_adjust_stats as usize);

    let stat_up: Vec<Stat> = (&mut rng)
        .sample_iter(Standard)
        .filter(|stat| adjust_stats.contains(stat))
        .take(total_adjusts)
        .collect();

    let stat_down: Vec<Stat> = rng
        .sample_iter::<Stat, Standard>(Standard)
        .filter(|stat| !adjust_stats.contains(stat))
        .take(total_adjusts)
        .collect();

    stat_up.iter().fold(StatBlock::default(), |x, y| x + *y)
        - stat_down.iter().fold(StatBlock::default(), |x, y| x + *y)
}

const NAMES: [&str; 20] = [
    "Garry", "Brad", "Rob", "Adrian", "Jeff", "Larry", "Chad", "Fred", "Jack", "Jerry", "Mitch",
    "Bob", "Billy", "Johnny", "Terry", "Berk", "Dave", "Steve", "Joe", "Bort",
];

pub fn random_character() -> Character {
    let mut rng = thread_rng();

    let class1 = rng.gen();

    create_character(
        NAMES.choose(&mut rng).unwrap().clone().into(),
        class1,
        (&mut rng)
            .sample_iter(Standard)
            .filter(|c| *c != class1)
            .next()
            .unwrap(),
        random_stat_choice(),
        random_stat_adjust(),
        rng.gen(),
        rng.gen(),
    )
}

#[cfg(test)]
mod random_character_tests {
    use super::*;

    #[test]
    fn test_choice() {
        for _i in 0..1000 {
            assert_eq!(Vec::from(random_stat_choice()).iter().sum::<i64>(), 2);
        }
    }

    #[test]
    fn test_adjust() {
        for _i in 0..1000 {
            let adjust = Vec::from(random_stat_adjust());
            println!("{:?}", adjust);
            assert_eq!(adjust.iter().sum::<i64>(), 0);
        }
    }

    #[test]
    fn character_stats() {
        for _i in 0..1000 {
            assert_eq!(
                Vec::from(random_character().stat_block).iter().sum::<i64>(),
                16
            );
        }
    }
}
