use std::{
    convert,
    ops::{self, Index},
};

use rand::{distributions::Standard, prelude::Distribution, seq::IteratorRandom};
use strum::IntoEnumIterator;
use strum_macros::EnumIter;

#[derive(Debug, Clone, Copy, PartialEq, Eq, EnumIter)]
pub enum Stat {
    STR,
    DEX,
    CON,
    STB,

    INT,
    KND,
    CMP,
    CHA,
}

impl ops::Add<Stat> for Stat {
    type Output = StatBlock;

    fn add(self, rhs: Stat) -> Self::Output {
        StatBlock::from(self) + rhs
    }
}

impl Distribution<Stat> for Standard {
    fn sample<R: rand::Rng + ?Sized>(&self, rng: &mut R) -> Stat {
        Stat::iter().choose(rng).unwrap()
    }
}

#[allow(non_snake_case)]
#[derive(Debug, PartialEq, Eq, Default, Clone, Copy)]
pub struct StatBlock {
    // Physical
    pub STR: i64,
    pub DEX: i64,
    pub CON: i64,
    pub STB: i64,
    // Mental
    pub INT: i64,
    pub KND: i64,
    pub CMP: i64,
    pub CHA: i64,
}

impl StatBlock {
    pub fn ones() -> StatBlock {
        StatBlock::from((1, 1, 1, 1, 1, 1, 1, 1))
    }
}

pub trait ToStatBlock {
    fn stat_block(&self) -> StatBlock;
}

impl ops::Add<StatBlock> for StatBlock {
    type Output = StatBlock;

    fn add(self, rhs: StatBlock) -> Self::Output {
        StatBlock {
            STR: self.STR + rhs.STR,
            DEX: self.DEX + rhs.DEX,
            CON: self.CON + rhs.CON,
            STB: self.STB + rhs.STB,
            INT: self.INT + rhs.INT,
            KND: self.KND + rhs.KND,
            CMP: self.CMP + rhs.CMP,
            CHA: self.CHA + rhs.CHA,
        }
    }
}

impl ops::Sub<StatBlock> for StatBlock {
    type Output = StatBlock;

    fn sub(self, rhs: StatBlock) -> Self::Output {
        self + -1 * rhs
    }
}

impl ops::Mul<StatBlock> for i64 {
    type Output = StatBlock;

    fn mul(self, rhs: StatBlock) -> Self::Output {
        StatBlock {
            STR: self * rhs.STR,
            DEX: self * rhs.DEX,
            CON: self * rhs.CON,
            STB: self * rhs.STB,
            INT: self * rhs.INT,
            KND: self * rhs.KND,
            CMP: self * rhs.CMP,
            CHA: self * rhs.CHA,
        }
    }
}

impl ops::Add<Stat> for StatBlock {
    type Output = StatBlock;

    fn add(self, rhs: Stat) -> Self::Output {
        self + StatBlock::from(rhs)
    }
}

impl convert::From<(i64, i64, i64, i64, i64, i64, i64, i64)> for StatBlock {
    #[allow(non_snake_case)]
    fn from(
        (STR, DEX, CON, STB, INT, KND, CMP, CHA): (i64, i64, i64, i64, i64, i64, i64, i64),
    ) -> Self {
        StatBlock {
            STR,
            DEX,
            CON,
            STB,
            INT,
            KND,
            CMP,
            CHA,
        }
    }
}

impl From<Stat> for StatBlock {
    fn from(value: Stat) -> Self {
        match value {
            Stat::STR => StatBlock::from((1, 0, 0, 0, 0, 0, 0, 0)),
            Stat::DEX => StatBlock::from((0, 1, 0, 0, 0, 0, 0, 0)),
            Stat::CON => StatBlock::from((0, 0, 1, 0, 0, 0, 0, 0)),
            Stat::STB => StatBlock::from((0, 0, 0, 1, 0, 0, 0, 0)),
            Stat::INT => StatBlock::from((0, 0, 0, 0, 1, 0, 0, 0)),
            Stat::KND => StatBlock::from((0, 0, 0, 0, 0, 1, 0, 0)),
            Stat::CMP => StatBlock::from((0, 0, 0, 0, 0, 0, 1, 0)),
            Stat::CHA => StatBlock::from((0, 0, 0, 0, 0, 0, 0, 1)),
        }
    }
}

impl Index<Stat> for StatBlock {
    type Output = i64;

    fn index(&self, index: Stat) -> &Self::Output {
        match index {
            Stat::STR => &self.STR,
            Stat::DEX => &self.DEX,
            Stat::CON => &self.CON,
            Stat::STB => &self.STB,
            Stat::INT => &self.INT,
            Stat::KND => &self.KND,
            Stat::CMP => &self.CMP,
            Stat::CHA => &self.CHA,
        }
    }
}

impl convert::From<StatBlock> for Vec<i64> {
    fn from(value: StatBlock) -> Self {
        vec![
            value.STR, value.DEX, value.CON, value.STB, value.INT, value.KND, value.CMP, value.CHA,
        ]
    }
}
